from __future__ import annotations

from typing import List, overload

from deciphon_intervals import Interval
from pydantic import BaseModel, RootModel

from deciphon_snap.fasta import FASTAItem, FASTAList
from deciphon_snap.gff import GFFItem, GFFList
from deciphon_snap.hit import HitList
from deciphon_snap.hmmer import H3Result
from deciphon_snap.match import LazyMatchList, Match, MatchElemName, MatchList
from deciphon_snap.query_interval import QueryIntervalBuilder

__all__ = ["Prod"]


class Prod(BaseModel):
    id: int
    seq_id: int
    window: int
    window_interval: Interval
    hit: int
    hit_interval: Interval
    profile: str
    abc: str
    lrt: float
    evalue: float
    match_list: LazyMatchList
    h3result: H3Result | None = None

    @property
    def gffs(self):
        gff_list = GFFList(root=[])
        for hit in self.hits:
            attr = f"Profile={self.profile};Alphabet={self.abc}"
            gff = GFFItem(
                seqid=str(self.seq_id),
                source="deciphon",
                type="CDS",
                start=hit.interval.r.start,
                end=hit.interval.r.stop,
                score=self.evalue,
                strand="+",
                phase="0",
                attributes=attr,
            )
            gff_list.root.append(gff)
        return gff_list

    @property
    def hits(self):
        offset = self.window_interval.py.start + self.hit_interval.py.start
        qibuilder = QueryIntervalBuilder(self.match_list.evaluate(), offset)
        hits = []
        for hit in HitList.make(self.match_list.evaluate()):
            hit.interval = qibuilder.make(hit.match_list_interval)
            hit.match_list = self.match_list.evaluate()
            hits.append(hit)
        return HitList.model_validate(hits)

    @property
    def matches(self):
        matches: list[Match] = []
        offset = self.window_interval.py.start + self.hit_interval.py.start
        for x in self.match_list:
            match = Match(raw=x.raw, start=x.start, end=x.end, position=offset)
            matches.append(match)
            offset += match.query_size
        return MatchList(root=matches)

    @property
    def hmmer(self):
        assert self.h3result is not None
        return self.h3result

    @property
    def query(self):
        return self.match_list.query

    @property
    def codon(self):
        return self.match_list.codon

    @property
    def amino(self):
        return self.match_list.amino


class ProdList(RootModel):
    root: List[Prod]

    def __len__(self):
        return len(self.root)

    @overload
    def __getitem__(self, i: int) -> Prod: ...

    @overload
    def __getitem__(self, i: slice) -> ProdList: ...

    def __getitem__(self, i: int | slice):
        if isinstance(i, slice):
            return ProdList.model_validate(self.root[i])
        prod = self.root[i]
        assert isinstance(prod, Prod)
        return prod

    def __iter__(self):
        return iter(self.root)

    def gff_list(self):
        return GFFList(root=[gff for prod in self.root for gff in prod.gffs])

    def fasta_list(self, name: MatchElemName):
        return FASTAList(root=list(self._fasta_list(name)))

    def _fasta_list(self, name: MatchElemName):
        for x in self.root:
            if name == MatchElemName.QUERY:
                yield FASTAItem(defline=str(x.seq_id), sequence=x.match_list.query)
            elif name == MatchElemName.STATE:
                yield FASTAItem(defline=str(x.seq_id), sequence=x.match_list.state)
            elif name == MatchElemName.CODON:
                yield FASTAItem(defline=str(x.seq_id), sequence=x.match_list.codon)
            elif name == MatchElemName.AMINO:
                yield FASTAItem(defline=str(x.seq_id), sequence=x.match_list.amino)
            else:
                assert False



from __future__ import annotations

from typing import List, overload

from pydantic import BaseModel, RootModel

__all__ = ["GFFItem", "GFFList"]


class GFFItem(BaseModel):
    seqid: str
    source: str
    type: str
    start: int
    end: int
    score: float
    strand: str
    phase: str
    attributes: str

    def format(self):
        score = f"{self.score:.2g}"
        return (
            f"{self.seqid}\t{self.source}\t{self.type}\t{self.start}\t{self.end}\t"
            + f"{score}\t{self.strand}\t{self.phase}\t{self.attributes}"
        )


class GFFList(RootModel):
    root: List[GFFItem]

    def __len__(self):
        return len(self.root)

    @overload
    def __getitem__(self, i: int) -> GFFItem: ...

    @overload
    def __getitem__(self, i: slice) -> GFFList: ...

    def __getitem__(self, i: int | slice):
        if isinstance(i, slice):
            return GFFList.model_validate(self.root[i])
        prod = self.root[i]
        assert isinstance(prod, GFFItem)
        return prod

    def __iter__(self):
        return iter(self.root)

    def format(self):
        gffs = [GFFItem.model_validate(x) for x in self.root]
        for i, x in enumerate(gffs):
            x.attributes = x.attributes + f";ID={i+1}"
        return "\n".join(["##gff-version 3"] + [x.format() for x in gffs]) + "\n"



from __future__ import annotations

import os
from io import StringIO
from tempfile import NamedTemporaryFile

from h3result.h3result import H3Result as H3ResultRaw
from hmmer_tables.domtbl import DomTBL, read_domtbl
from hmmer_tables.tbl import TBL, read_tbl
from pydantic import BaseModel, ConfigDict

__all__ = ["H3Result"]


class H3Result(BaseModel):
    raw: H3ResultRaw
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def targets(self) -> str:
        tmp = TempFile()
        with tmp:
            self.raw.print_targets(tmp.fileno())
        return tmp.content

    @property
    def domains(self) -> str:
        tmp = TempFile()
        with tmp:
            self.raw.print_domains(tmp.fileno())
        return tmp.content

    @property
    def targets_table(self) -> str:
        tmp = TempFile()
        with tmp:
            self.raw.print_targets_table(tmp.fileno())
        return tmp.content

    @property
    def domains_table(self) -> str:
        tmp = TempFile()
        with tmp:
            self.raw.print_domains_table(tmp.fileno())
        return tmp.content

    @property
    def tbl(self) -> TBL:
        return read_tbl(stream=StringIO(self.targets_table))

    @property
    def domtbl(self) -> DomTBL:
        return read_domtbl(stream=StringIO(self.domains_table))

    def __str__(self):
        items = []
        for x in self.domtbl:
            e_value = x.full_sequence.e_value
            start = x.ali_coord.start
            stop = x.ali_coord.stop
            items.append(f"({e_value},{start}..{stop})")
        return "[{}]".format(",".join(items))

    def __repr__(self):
        return self.__str__()


class TempFile:
    def __init__(self):
        self._content = None
        self._stream = None

    def fileno(self):
        assert self._stream is not None
        return self._stream.fileno()

    @property
    def content(self):
        assert isinstance(self._content, str)
        return self._content

    def __enter__(self):
        self._stream = NamedTemporaryFile(delete=False)

    def __exit__(self, *args, **kwargs):
        del args
        del kwargs
        assert self._stream is not None
        self._stream.close()
        self._content = open(self._stream.name).read()
        os.unlink(self._stream.name)



from fsspec.implementations.cached import SimpleCacheFileSystem
from fsspec.implementations.memory import MemoryFile
from fsspec.implementations.zip import ZipFileSystem

from deciphon_snap.path_like import PathLike
from deciphon_snap.snap_file import SnapFile

__all__ = ["read_snap"]


def read_snap(filename: PathLike):
    fo = filename.decode() if isinstance(filename, bytes) else str(filename)
    with open(fo, "rb") as f:
        fs = ZipFileSystem(fo=MemoryFile(data=f.read()))
        return SnapFile(SimpleCacheFileSystem(fs=fs))



from deciphon_intervals import PyInterval

from deciphon_snap.match import MatchList, MatchListInterval

__all__ = ["QueryInterval", "QueryIntervalBuilder"]


class QueryInterval(PyInterval): ...


class QueryIntervalBuilder:
    def __init__(self, match_list: MatchList, offset: int):
        self._offset = []
        for x in match_list:
            self._offset.append(offset)
            offset += x.query_size
        self._offset.append(offset)

    def make(self, match_list_interval: MatchListInterval) -> QueryInterval:
        i = match_list_interval
        start = self._offset[i.py.start]
        stop = self._offset[i.py.stop]
        return QueryInterval(start=start, stop=stop)



from __future__ import annotations

from io import StringIO
from typing import List, overload

from fasta_reader.writer import Writer
from pydantic import BaseModel, RootModel

__all__ = ["FASTAItem", "FASTAList"]


class FASTAItem(BaseModel):
    defline: str
    sequence: str


class FASTAList(RootModel):
    root: List[FASTAItem]

    def __len__(self):
        return len(self.root)

    @overload
    def __getitem__(self, i: int) -> FASTAItem: ...

    @overload
    def __getitem__(self, i: slice) -> FASTAList: ...

    def __getitem__(self, i: int | slice):
        if isinstance(i, slice):
            return FASTAList.model_validate(self.root[i])
        prod = self.root[i]
        assert isinstance(prod, FASTAItem)
        return prod

    def __iter__(self):
        return iter(self.root)

    def format(self, ncols: int = 60):
        t = StringIO()
        w = Writer(t, ncols=ncols)
        for x in self.root:
            w.write_item(x.defline, x.sequence)
        return t.getvalue()



from __future__ import annotations

from typing import List, Type, TypeVar, overload

from pydantic import BaseModel, RootModel

from deciphon_snap.match import Match, MatchList, MatchListInterval
from deciphon_snap.query_interval import QueryInterval

__all__ = ["Hit", "HitList"]


class Hit(BaseModel):
    id: int
    match_list_interval: MatchListInterval
    _interval: QueryInterval | None = None
    _match_list: MatchList | None = None

    @property
    def interval(self):
        assert self._interval is not None
        return self._interval

    @interval.setter
    def interval(self, x: QueryInterval):
        self._interval = x

    @property
    def match_list(self):
        assert self._match_list is not None
        return self._match_list

    @match_list.setter
    def match_list(self, x: MatchList):
        self._match_list = x

    @property
    def matches(self):
        assert self._interval is not None
        assert self._match_list is not None
        matches: list[Match] = []
        offset = self._interval.py.start
        for x in self._match_list[self.match_list_interval.slice]:
            y = Match(raw=x.raw, start=x.start, end=x.end, position=offset)
            if y.is_match_state or y.is_insert_state:
                offset += y.query_size
            matches.append(y)
        return matches


T = TypeVar("T", bound="HitList")


class HitList(RootModel):
    root: List[Hit]

    def __len__(self):
        return len(self.root)

    @overload
    def __getitem__(self, i: int) -> Hit: ...

    @overload
    def __getitem__(self, i: slice) -> HitList: ...

    def __getitem__(self, i: int | slice):
        if isinstance(i, slice):
            return HitList.model_validate(self.root[i])
        hit = self.root[i]
        assert isinstance(hit, Hit)
        return hit

    def __iter__(self):
        return iter(self.root)

    def __str__(self):
        return " ".join(str(i) for i in self.root)

    @classmethod
    def make(cls: Type[T], match_list: MatchList) -> T:
        hits: List[Hit] = []

        offset = 0
        hit_start_found = False
        hit_end_found = False
        match_start = 0
        match_stop = 0

        for i, x in enumerate(match_list):
            if not hit_start_found and x.is_core_state:
                match_start = i
                hit_start_found = True

            if hit_start_found and not x.is_core_state:
                hit_end_found = True

            if hit_end_found:
                match_stop = i
                mi = MatchListInterval(start=match_start, stop=match_stop)
                hit_id = len(hits)
                hit = Hit(id=hit_id, match_list_interval=mi)
                hits.append(hit)
                hit_start_found = False
                hit_end_found = False

            offset += x.query_size

        return cls.model_validate(hits)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from typing import List, overload

from deciphon_intervals import PyInterval
from pydantic import BaseModel, ConfigDict

from deciphon_snap.amino import AminoInterval

__all__ = ["Match", "MatchList", "LazyMatchList", "MatchListInterval", "MatchElemName"]


class MatchElemName(Enum):
    QUERY = 1
    STATE = 2
    CODON = 3
    AMINO = 4


@dataclass(slots=True, frozen=True)
class Match:
    raw: str
    start: int
    end: int
    position: int = -1

    @classmethod
    def from_string(cls, x: str):
        return cls(raw=x, start=0, end=len(x))

    @property
    def query(self):
        start = self.start
        return self.raw[start : self.raw.find(",", start, self.end)]

    @property
    def state(self):
        i = self.start
        i = self.raw.find(",", i, self.end) + 1
        return self.raw[i : self.raw.find(",", i, self.end)]

    @property
    def codon(self):
        i = self.start
        i = self.raw.find(",", i, self.end) + 1
        i = self.raw.find(",", i, self.end) + 1
        return self.raw[i : self.raw.find(",", i, self.end)]

    @property
    def amino(self):
        i = self.start
        i = self.raw.find(",", i, self.end) + 1
        i = self.raw.find(",", i, self.end) + 1
        i = self.raw.find(",", i, self.end) + 1
        return self.raw[i : self.end]

    @property
    def query_size(self) -> int:
        return self.raw.find(",", self.start, self.end) - self.start

    @property
    def _state_symbol(self):
        i = self.start
        i = self.raw.find(",", i, self.end) + 1
        return self.raw[i]

    @property
    def is_insert_state(self):
        return self._state_symbol == "I"

    @property
    def is_match_state(self):
        return self._state_symbol == "M"

    @property
    def is_delete_state(self):
        return self._state_symbol == "D"

    @property
    def is_core_state(self):
        x = self._state_symbol
        return x == "I" or x == "M" or x == "D"

    def __str__(self):
        query = self.query if len(self.query) > 0 else "∅"
        state = self.state
        codon = self.codon if len(self.codon) > 0 else "∅"
        amino = self.amino if len(self.amino) > 0 else "∅"
        return f"({query},{state},{codon},{amino})"


@dataclass(slots=True, frozen=True)
class MatchList:
    root: List[Match]

    @classmethod
    def from_string(cls, x: str):
        y = [i for i in ifind(x, ";")]
        return cls([Match(raw=x, start=i[0], end=i[1]) for i in y])

    def __len__(self):
        return len(self.root)

    @overload
    def __getitem__(self, i: int) -> Match: ...

    @overload
    def __getitem__(self, i: slice) -> MatchList: ...

    def __getitem__(self, i: int | slice):
        if isinstance(i, slice):
            return MatchList(self.root[i])
        match = self.root[i]
        assert isinstance(match, Match)
        return match

    def __iter__(self):
        return iter(self.root)

    def __str__(self):
        return " ".join(str(i) for i in self.root)

    @property
    def query(self):
        return "".join(x.query for x in iter(self))

    @property
    def state(self):
        return "".join(x.state for x in iter(self))

    @property
    def codon(self):
        return "".join(x.codon for x in iter(self))

    @property
    def amino(self):
        return "".join(x.amino for x in iter(self))


class MatchListInterval(PyInterval): ...


class MatchListIntervalBuilder:
    def __init__(self, match_list: MatchList):
        self._amino_map = [i for i, x in enumerate(match_list) if len(x.amino) > 0]

    def make_from_amino_interval(
        self, amino_interval: AminoInterval
    ) -> MatchListInterval:
        i = amino_interval
        x = self._amino_map[i.slice]
        return MatchListInterval(start=x[0], stop=x[-1] + 1)


class LazyMatchList(BaseModel):
    raw: str
    model_config = ConfigDict(frozen=True)

    @lru_cache(maxsize=1)
    def evaluate(self):
        return MatchList.from_string(self.raw)

    def __len__(self):
        return len(self.evaluate())

    def __getitem__(self, i):
        return self.evaluate()[i]

    def __iter__(self):
        return iter(self.evaluate())

    def __str__(self):
        return str(self.evaluate())

    def __repr__(self):
        return repr(self.evaluate())

    @property
    def query(self):
        return self.evaluate().query

    @property
    def state(self):
        return self.evaluate().state

    @property
    def codon(self):
        return self.evaluate().codon

    @property
    def amino(self):
        return self.evaluate().amino


def ifind(x: str, delim: str):
    start = 0
    end = x.find(delim, start)
    while end != -1:
        yield (start, end)
        start = end + 1
        end = x.find(delim, start)
    yield (start, len(x))



from functools import partial
from io import StringIO
from itertools import accumulate, chain
from typing import Optional

from hmmer_tables.query import DomAnnot, read_query
from pydantic import BaseModel
from tabulate import simple_separated_format, tabulate

from deciphon_snap.match import Match, MatchList
from deciphon_snap.prod import H3Result, Prod
from deciphon_snap.snap_file import SnapFile

__all__ = ["view_alignments"]


class DeciStep(BaseModel):
    match: Match


class HMMERStep(BaseModel):
    hmm_pos: int
    hmm_cs: str
    amino_pos: int
    amino_cs: str
    amino: str
    match: str
    score: str


class AssocStep(BaseModel):
    deci: Optional[DeciStep]
    hmmer: Optional[HMMERStep]


def make_deciphon_steps(match_list: MatchList):
    return list(enumerate([DeciStep(match=x) for x in match_list if len(x.amino) > 0]))


def make_hmmer_annot(h3result: H3Result):
    hmmer_query = read_query(stream=StringIO(h3result.domains))
    assert len(hmmer_query.domains) == 1
    return hmmer_query.domains[0]


def make_hmmer_steps(annot: DomAnnot):
    hmmer: list[tuple[int, HMMERStep]] = []
    for x in annot.aligns:
        a = x.align
        core_positions = a.core_positions
        core_positions = list(range(a.core_interval.start, a.core_interval.stop + 1))
        start = a.query_interval.py.start
        stop = a.query_interval.py.stop
        offset = start - 1
        amino_pos = [offset + i for i in accumulate([int(i != "-") for i in a.query])]
        assert amino_pos[0] == start
        assert amino_pos[-1] + 1 == stop
        for i in range(len(core_positions)):
            step = HMMERStep(
                hmm_pos=core_positions[i],
                hmm_cs=a.hmm_cs[i],
                amino_pos=amino_pos[i],
                amino_cs=a.query_cs[i],
                amino=a.query[i],
                match=a.match[i],
                score=a.score[i],
            )
            hmmer.append((amino_pos[i], step))
    return hmmer


def assoc_steps(deci: list[tuple[int, DeciStep]], hmmer: list[tuple[int, HMMERStep]]):
    steps: list[AssocStep] = []
    di = 0
    hi = 0
    while di < len(deci) and hi < len(hmmer):
        deci_pos = deci[di][0]
        hmmer_pos = hmmer[hi][0]

        if deci_pos < hmmer_pos:
            steps.append(AssocStep(deci=deci[di][1], hmmer=None))
            di += 1

        if hmmer_pos < deci_pos:
            steps.append(AssocStep(deci=None, hmmer=hmmer[hi][1]))
            hi += 1

        if deci_pos == hmmer_pos:
            steps.append(AssocStep(deci=deci[di][1], hmmer=hmmer[hi][1]))
            di += 1
            hi += 1

    while di < len(deci):
        steps.append(AssocStep(deci=deci[di][1], hmmer=None))
        di += 1

    while hi < len(hmmer):
        steps.append(AssocStep(deci=None, hmmer=hmmer[hi][1]))
        hi += 1

    return steps


def flat(x):
    return list(chain.from_iterable(x))


def view_alignment(prod: Prod):
    assert prod.h3result is not None
    annot = make_hmmer_annot(prod.h3result)
    txt = "Alignments for each domain:\n"
    if len(annot.aligns) == 0:
        txt += annot.head.strip() + "\n"
        return txt

    deci = make_deciphon_steps(prod.matches)
    hmmer = make_hmmer_steps(annot)

    align_head = annot.aligns[0].head
    txt += align_head.strip() + "\n"

    profile = annot.aligns[0].align.profile
    query_name = annot.aligns[0].align.query_name

    steps: list[AssocStep] = assoc_steps(deci, hmmer)

    def grab_amino(x):
        if x.hmmer is not None and x.deci is not None:
            return x.deci.match.amino
        if x.hmmer is not None and x.deci is None:
            return "-"
        return ""

    def grab_query(x, i):
        if x.hmmer is not None and x.deci is not None:
            q = x.deci.match.query
            return q[i] if len(q) > i else "."
        if x.hmmer is not None and x.deci is None:
            return "-"
        return ""

    query_pos = []
    curr_pos = 0
    for x in steps:
        if x.hmmer is not None and x.deci is not None:
            curr_pos = x.deci.match.position
            query_pos.append(curr_pos)
            curr_pos += len(x.deci.match.query)
        if x.hmmer is not None and x.deci is None:
            query_pos.append(curr_pos)

    hmm_cs = flat(("" if x.hmmer is None else x.hmmer.hmm_cs) for x in steps)
    query = flat(("" if x.hmmer is None else x.hmmer.amino) for x in steps)
    match = flat(("" if x.hmmer is None else x.hmmer.match) for x in steps)

    amino = flat(grab_amino(x) for x in steps)
    q0 = flat(partial(grab_query, i=0)(x) for x in steps)
    q1 = flat(partial(grab_query, i=1)(x) for x in steps)
    q2 = flat(partial(grab_query, i=2)(x) for x in steps)
    q3 = flat(partial(grab_query, i=3)(x) for x in steps)
    q4 = flat(partial(grab_query, i=4)(x) for x in steps)
    score = flat(("" if x.hmmer is None else x.hmmer.score) for x in steps)

    hmm_pos = []
    for x in steps:
        if x.hmmer is not None:
            hmm_pos.append(x.hmmer.hmm_pos)

    amino_pos = []
    for x in steps:
        if x.hmmer is not None:
            amino_pos.append(x.hmmer.amino_pos)

    table = []
    n = len(hmm_cs)
    COLS = 96
    width = min(COLS, n)
    for i in range(0, n, COLS):
        sl = slice(i, min(i + COLS, n))
        pad = "&" * (width - (sl.stop - sl.start))
        hmm_pos_left = hmm_pos[sl][0]
        hmm_pos_right = hmm_pos[sl][-1]
        amino_pos_left = amino_pos[sl][0] + 1
        amino_pos_right = amino_pos[sl][-1] + 1
        query_pos_left = query_pos[sl][0] + 1
        query_pos_right = query_pos[sl][-1] + 1
        row = [
            [None, None, "".join(hmm_cs[sl]) + pad, "CS"],
            [profile, hmm_pos_left, "".join(query[sl]) + pad, hmm_pos_right],
            [None, None, "".join(match[sl]) + pad, None],
            [None, amino_pos_left, "".join(amino[sl]) + pad, amino_pos_right],
            [query_name, query_pos_left, "".join(q0[sl]) + pad, query_pos_right],
            [None, None, "".join(q1[sl]) + pad, None],
            [None, None, "".join(q2[sl]) + pad, None],
            [None, None, "".join(q3[sl]) + pad, None],
            [None, None, "".join(q4[sl]) + pad, None],
            [None, None, "".join(score[sl]) + pad, "PP"],
        ]
        table += row + [[None, None, None, None]]
    tablefmt = simple_separated_format(" ")
    txt += tabulate(
        table, tablefmt=tablefmt, colalign=("right", "right", "left", "left")
    )
    txt = txt.replace("&", "") + "\n"
    return txt


def view_alignments(snap: SnapFile):
    txt = []
    for prod in snap.products:
        txt.append(view_alignment(prod))
    return "\n".join(txt)
